// license:BSD-3-Clause
// copyright-holders:Olivier Galibert

#ifndef MAME_BUS_WAVEBLASTER_DB50XG_H
#define MAME_BUS_WAVEBLASTER_DB50XG_H

// Yamaha DB50XG

#pragma once

#include "waveblaster.h"

DECLARE_DEVICE_TYPE(DB50XG, device_waveblaster_interface)

#endif // MAME_BUS_WAVEBLASTER_DB50XG_H
