// license:BSD-3-Clause
// copyright-holders:Vas Crabb
/***********************************************************************

    NEC PC Engine/TurboGrafx-16 Mouse emulation

***********************************************************************/
#ifndef MAME_BUS_PCE_CTRL_MOUSE_H
#define MAME_BUS_PCE_CTRL_MOUSE_H

#pragma once

#include "pcectrl.h"


DECLARE_DEVICE_TYPE(PCE_MOUSE, device_pce_control_port_interface)

#endif // MAME_BUS_PCE_CTRL_MOUSE_H
