// license:BSD-3-Clause
// copyright-holders:Sergey Svishchev
/***************************************************************************

    HP-HIL devices

***************************************************************************/

#ifndef MAME_BUS_HP_HIL_HIL_DEVICES_H
#define MAME_BUS_HP_HIL_HIL_DEVICES_H


#define STR_KBD_HP_46020A    "hp_46020a"    // ITF Keyboard
#define STR_KBD_HP_46021A    "hp_46021a"    // ITF Keyboard
#define STR_KBD_HP_46030A    "hp_46030a"    // Vectra Keyboard
#define STR_KBD_HP_INTEGRAL  "hp_ipc_kbd"   // Integral Keyboard

#define STR_MOUSE_HP_46060A    "hp_46060a"  // 2-button mouse
#define STR_MOUSE_HP_46060B    "hp_46060b"  // 3-button mouse


void hp_hil_devices(device_slot_interface &device);


#endif // MAME_BUS_HP_HIL_HIL_DEVICES_H
