// license:BSD-3-Clause
// copyright-holders:Vas Crabb
#ifndef MAME_BUS_INTELLEC4_TAPEREADER_H
#define MAME_BUS_INTELLEC4_TAPEREADER_H

#pragma once

#include "intellec4.h"

DECLARE_DEVICE_TYPE_NS(INTELLEC4_TAPE_READER, bus::intellec4, device_univ_card_interface)

#endif // MAME_BUS_INTELLEC4_TAPEREADER_H
