// license: GPL-2.0+
// copyright-holders: Dirk Best
/***************************************************************************

    SAM Coupe Expansion Slot modules

***************************************************************************/

#ifndef MAME_BUS_SAMCOUPE_EXPANSION_MODULES_H
#define MAME_BUS_SAMCOUPE_EXPANSION_MODULES_H

#pragma once

void samcoupe_expansion_modules(device_slot_interface &device);

#endif // MAME_BUS_SAMCOUPE_EXPANSION_MODULES_H
