// license:BSD-3-Clause
// copyright-holders:Wilbert Pol
#ifndef MAME_BUS_MSX_CART_RTYPE_H
#define MAME_BUS_MSX_CART_RTYPE_H

#pragma once

#include "bus/msx/slot/cartridge.h"


DECLARE_DEVICE_TYPE(MSX_CART_RTYPE, msx_cart_interface)

#endif // MAME_BUS_MSX_CART_RTYPE_H
