/*
 * Copyright 2010-2022 Branimir Karadzic. All rights reserved.
 * License: https://github.com/bkaradzic/bx/blob/master/LICENSE
 */

#ifndef BX_TIMER_H_HEADER_GUARD
#define BX_TIMER_H_HEADER_GUARD

#include "bx.h"

namespace bx
{
	///
	int64_t getHPCounter();

	///
	int64_t getHPFrequency();

} // namespace bx

#endif // BX_TIMER_H_HEADER_GUARD
