// license:BSD-3-Clause
// copyright-holders:R. Justice
/*********************************************************************

    vistaa800.h

    Vista A800 8" disk Controller Card

*********************************************************************/

#ifndef MAME_BUS_A2BUS_VISTAA800_H
#define MAME_BUS_A2BUS_VISTAA800_H

#pragma once

#include "a2bus.h"

DECLARE_DEVICE_TYPE(A2BUS_VISTAA800, device_a2bus_card_interface)

#endif  // MAME_BUS_A2BUS_VISTAA800_H
