// license:BSD-3-Clause
// copyright-holders:R. Belmont
/*********************************************************************

    a2applicard.h

    Implementation of the PCPI AppliCard Z-80 card

*********************************************************************/

#ifndef MAME_BUS_A2BUS_A2APPLICARD_H
#define MAME_BUS_A2BUS_A2APPLICARD_H

#pragma once

#include "a2bus.h"

// device type declaration
DECLARE_DEVICE_TYPE(A2BUS_APPLICARD, device_a2bus_card_interface)

#endif // MAME_BUS_A2BUS_A2APPLICARD_H
