// license:BSD-3-Clause
// copyright-holders:Nathan Woods
#ifndef MAME_BUS_COCO_COCO_ORCH90_H
#define MAME_BUS_COCO_COCO_ORCH90_H

#pragma once

#include "cococart.h"

// device type definition
DECLARE_DEVICE_TYPE(COCO_ORCH90, device_cococart_interface)

#endif // MAME_BUS_COCO_COCO_ORCH90_H
