// license:BSD-3-Clause
// copyright-holders:Nigel Barnes
#ifndef MAME_BUS_PSION_HONDA_PARALLEL_H
#define MAME_BUS_PSION_HONDA_PARALLEL_H

#pragma once

#include "slot.h"


DECLARE_DEVICE_TYPE(PSION_PARALLEL, device_psion_honda_interface)

#endif // MAME_BUS_PSION_HONDA_PARALLEL_H
