// license:BSD-3-Clause
// copyright-holders:R. Belmont
/*********************************************************************

    pc_xporter.h

    Implementation of the Applied Engineering PC Transporter card

*********************************************************************/

#ifndef MAME_BUS_A2BUS_PC_XPORTER_H
#define MAME_BUS_A2BUS_PC_XPORTER_H

#pragma once

#include "a2bus.h"

// device type declaration
DECLARE_DEVICE_TYPE(A2BUS_PCXPORTER, device_a2bus_card_interface)

#endif // MAME_BUS_A2BUS_PC_XPORTER_H
