// license:BSD-3-Clause
// copyright-holders:R. Belmont, Vas Crabb
#ifndef MAME_BUS_NUBUS_NUBUS_SPEC8_H
#define MAME_BUS_NUBUS_NUBUS_SPEC8_H

#pragma once

#include "nubus.h"

// device type definition
DECLARE_DEVICE_TYPE(NUBUS_SPEC8S3, device_nubus_card_interface)

#endif // MAME_BUS_NUBUS_NUBUS_SPEC8_H
