// license:BSD-3-Clause
// copyright-holders:Nigel Barnes
#ifndef MAME_BUS_PSION_HONDA_SSD_H
#define MAME_BUS_PSION_HONDA_SSD_H


#include "slot.h"


DECLARE_DEVICE_TYPE(PSION_SIENA_SSD, device_psion_honda_interface)

#endif // MAME_BUS_PSION_HONDA_SSD_H
