// license:BSD-3-Clause
// copyright-holders:Wilbert Pol, Fabio Priuli
/******************************************************************************

    Magnavox The Voice emulation

******************************************************************************/

#ifndef MAME_BUS_ODYSSEY2_VOICE_H
#define MAME_BUS_ODYSSEY2_VOICE_H

#pragma once

#include "slot.h"

DECLARE_DEVICE_TYPE(O2_ROM_VOICE, device_o2_cart_interface)

#endif // MAME_BUS_ODYSSEY2_VOICE_H
