// license:BSD-3-Clause
// copyright-holders:hap
/*******************************************************************************

    Saitek OSA Module: Kasparov Sparc

*******************************************************************************/

#ifndef MAME_BUS_SAITEK_OSA_SPARC_H
#define MAME_BUS_SAITEK_OSA_SPARC_H

#pragma once

#include "expansion.h"

DECLARE_DEVICE_TYPE(OSA_SPARC, device_saitekosa_expansion_interface)

#endif // MAME_BUS_SAITEK_OSA_SPARC_H
