// license:BSD-3-Clause
// copyright-holders:tim lindner
#ifndef MAME_BUS_MC10_PAK_H
#define MAME_BUS_MC10_PAK_H

#pragma once

#include "mc10_cart.h"

DECLARE_DEVICE_TYPE(MC10_PAK, device_mc10cart_interface)

#endif // MAME_BUS_MC10_PAK_H
