// license:BSD-3-Clause
// copyright-holders:R. Belmont
/*********************************************************************

    4play.h

    4 player digital joystick card for the Apple II by Lukazi
    http://lukazi.blogspot.com/2016/05/apple-ii-4play-joystick-card-revb.html

*********************************************************************/

#ifndef MAME_BUS_A2BUS_4PLAY_H
#define MAME_BUS_A2BUS_4PLAY_H

#include "a2bus.h"

// device type definition
DECLARE_DEVICE_TYPE(A2BUS_4PLAY, device_a2bus_card_interface)

#endif // MAME_BUS_A2BUS_4PLAY_H
