// license:BSD-3-Clause
// copyright-holders: Devin Acker

#ifndef MAME_BUS_WAVEBLASTER_WG130_H
#define MAME_BUS_WAVEBLASTER_WG130_H

// Casio WG130

#pragma once

#include "waveblaster.h"

DECLARE_DEVICE_TYPE(WG130, device_waveblaster_interface)

#endif // MAME_BUS_WAVEBLASTER_WG130_H
