// license:BSD-3-Clause
// copyright-holders:AJR
/**********************************************************************

    PolyMorphic Systems Video Terminal Interface

**********************************************************************/

#ifndef MAME_BUS_S100_POLYVTI_H
#define MAME_BUS_S100_POLYVTI_H

#pragma once

#include "bus/s100/s100.h"

DECLARE_DEVICE_TYPE(S100_POLY_VTI, device_s100_card_interface)

#endif // MAME_BUS_S100_POLYVTI_H
