// license:BSD-3-Clause
// copyright-holders:Sven Schnelle

#ifndef MAME_BUS_HP_DIO_HP98624_H
#define MAME_BUS_HP_DIO_HP98624_H

#pragma once

#include "hp_dio.h"

DECLARE_DEVICE_TYPE_NS(HPDIO_98624, bus::hp_dio, device_dio16_card_interface)

#endif // MAME_BUS_HP_DIO_HP98624_H
