// license:BSD-3-Clause
// copyright-holders:R. Belmont
/*********************************************************************

    a2dx1.h

    Implementation of the Decillionix DX-1 sampler card

*********************************************************************/

#ifndef MAME_BUS_A2BUS_A2DX1_H
#define MAME_BUS_A2BUS_A2DX1_H

#pragma once

#include "a2bus.h"

// device type declaration
DECLARE_DEVICE_TYPE(A2BUS_DX1, device_a2bus_card_interface)

#endif // MAME_BUS_A2BUS_A2DX1_H
