// license:BSD-3-Clause
// copyright-holders:Mark Garlanger
/***************************************************************************

  SigmaSoft Sound Effects Board

****************************************************************************/

#ifndef MAME_BUS_HEATHZENITH_H89_SIGMASOFT_SOUND_H
#define MAME_BUS_HEATHZENITH_H89_SIGMASOFT_SOUND_H

#pragma once

#include "h89bus.h"

DECLARE_DEVICE_TYPE(H89BUS_SIGMASOFT_SND, device_h89bus_right_card_interface)

#endif // MAME_BUS_HEATHZENITH_H89_SIGMASOFT_SOUND_H
