// license:BSD-3-Clause
// copyright-holders:Nigel Barnes
#ifndef MAME_BUS_BBC_1MHZBUS_BARRYBOX_H
#define MAME_BUS_BBC_1MHZBUS_BARRYBOX_H

#include "1mhzbus.h"


DECLARE_DEVICE_TYPE(BBC_BARRYBOX, device_bbc_1mhzbus_interface);

#endif // MAME_BUS_BBC_1MHZBUS_BARRYBOX_H
