// license:BSD-3-Clause
// copyright-holders:F. Ulivi

/*********************************************************************

    HP98628 Data communication interface
    HP98629 SRM interface

*********************************************************************/

#ifndef MAME_BUS_HP_DIO_HP98628_9_H
#define MAME_BUS_HP_DIO_HP98628_9_H

#pragma once

#include "hp_dio.h"

DECLARE_DEVICE_TYPE_NS(HPDIO_98628, bus::hp_dio, device_dio16_card_interface)
DECLARE_DEVICE_TYPE_NS(HPDIO_98629, bus::hp_dio, device_dio16_card_interface)

#endif /* MAME_BUS_HP_DIO_HP98628_9_H */
