// license:BSD-3-Clause
// copyright-holders:Curt Coder
#ifndef MAME_BUS_RS232_TELETEX800_H
#define MAME_BUS_RS232_TELETEX800_H

#pragma once

#include "rs232.h"


DECLARE_DEVICE_TYPE(TELETEX_800, device_rs232_port_interface)

#endif // MAME_BUS_RS232_TELETEX800_H
