// license:BSD-3-Clause
// copyright-holders:Wilbert Pol
#ifndef MAME_BUS_MSX_CART_DISK_H
#define MAME_BUS_MSX_CART_DISK_H

#pragma once

void msx_cart_disk_register_options(device_slot_interface &device);

#endif // MAME_BUS_MSX_CART_DISK_H
