// license:BSD-3-Clause
// copyright-holders:Aaron Giles

void fd1094_regenerate_key(running_machine &machine);
void fd1094_init_debugging(running_machine &machine, const char *cpureg, const char *keyreg, const char *statreg, void (*changed)(running_machine &));
