// license: GPL-2.0+
// copyright-holders: Dirk Best
/***************************************************************************

    BennVenn SD Loader for VZ300

***************************************************************************/

#ifndef MAME_BUS_VTECH_MEMEXP_SDLOADER_H
#define MAME_BUS_VTECH_MEMEXP_SDLOADER_H

#pragma once

#include "memexp.h"


DECLARE_DEVICE_TYPE(VTECH_SDLOADER, vtech_memexp_device)

#endif // MAME_BUS_VTECH_MEMEXP_SDLOADER_H
