// license:BSD-3-Clause
// copyright-holders:Dirk Best
/**********************************************************************

    ISA cards

**********************************************************************/

#ifndef MAME_BUS_ISA_ISA_CARDS_H
#define MAME_BUS_ISA_ISA_CARDS_H

#pragma once


// supported devices
void pc_isa8_cards(device_slot_interface &device);
void pc_isa16_cards(device_slot_interface &device);

#endif // MAME_BUS_ISA_ISA_CARDS_H
