// license:BSD-3-Clause
// copyright-holders:Nigel Barnes
/**********************************************************************

    Acorn AKA25 Ethernet

**********************************************************************/

#ifndef MAME_BUS_ARCHIMEDES_PODULE_ETHER1_H
#define MAME_BUS_ARCHIMEDES_PODULE_ETHER1_H

#pragma once

#include "slot.h"

// device type definition
DECLARE_DEVICE_TYPE(ARC_ETHER1_AKA25, device_archimedes_podule_interface)

#endif // MAME_BUS_ARCHIMEDES_PODULE_ETHER1_H
