// license:BSD-3-Clause
// copyright-holders:Nigel Barnes
#ifndef MAME_BUS_MPF1_VID_H
#define MAME_BUS_MPF1_VID_H

#pragma once

#include "slot.h"


DECLARE_DEVICE_TYPE(MPF_VID, device_mpf1_exp_interface)

#endif // MAME_BUS_MPF1_VID_H
