// license:BSD-3-Clause
// copyright-holders:hap
/**********************************************************************

    Sony Magic Key emulation

**********************************************************************/

#ifndef MAME_BUS_MSX_CTRL_MAGICKEY_H
#define MAME_BUS_MSX_CTRL_MAGICKEY_H

#pragma once

#include "ctrl.h"


DECLARE_DEVICE_TYPE(MSX_MAGICKEY, device_msx_general_purpose_port_interface)

#endif // MAME_BUS_MSX_CTRL_MAGICKEY_H
