// license:BSD-3-Clause
// copyright-holders:hap
/**********************************************************************

    Videopac+ C7420 Home Computer Module emulation

**********************************************************************/

#ifndef MAME_BUS_ODYSSEY2_HOMECOMP_H
#define MAME_BUS_ODYSSEY2_HOMECOMP_H

#pragma once

#include "slot.h"

DECLARE_DEVICE_TYPE(O2_ROM_HOMECOMP, device_o2_cart_interface)

#endif // MAME_BUS_ODYSSEY2_HOMECOMP_H
