// license:BSD-3-Clause
// copyright-holders:Andrei I. Holub

#ifndef MAME_BUS_SPECTRUM_NEOGS_H
#define MAME_BUS_SPECTRUM_NEOGS_H

#pragma once

#include "zxbus.h"


DECLARE_DEVICE_TYPE(ZXBUS_NEOGS, device_zxbus_card_interface)

#endif // MAME_BUS_SPECTRUM_NEOGS_H
