// license:BSD-3-Clause
// copyright-holders:Fabio Priuli
/**********************************************************************

    Sega Mark III "Joypad" / Master System "Control Pad" emulation

**********************************************************************/
#ifndef MAME_BUS_SMS_CTRL_JOYPAD_H
#define MAME_BUS_SMS_CTRL_JOYPAD_H

#pragma once

#include "smsctrl.h"


DECLARE_DEVICE_TYPE(SMS_JOYPAD, device_sms_control_interface)

#endif // MAME_BUS_SMS_CTRL_JOYPAD_H
