// license:BSD-3-Clause
// copyright-holders:R. Belmont
/*********************************************************************

    mouse.h

    Implemention of the Apple II Mouse Card

*********************************************************************/

#ifndef MAME_BUS_A2BUS_MOUSE_H
#define MAME_BUS_A2BUS_MOUSE_H

#pragma once

#include "a2bus.h"

DECLARE_DEVICE_TYPE(A2BUS_MOUSE, device_a2bus_card_interface)

#endif // MAME_BUS_A2BUS_MOUSE_H
