// license:BSD-3-Clause
// copyright-holders:R. Belmont
#ifndef MAME_BUS_NUBUS_PDS30_30HR_H
#define MAME_BUS_NUBUS_PDS30_30HR_H

#pragma once

#include "nubus.h"


DECLARE_DEVICE_TYPE(PDS030_XCEED30HR, device_nubus_card_interface)
DECLARE_DEVICE_TYPE(PDS030_XCEED30HRINT, device_nubus_card_interface)
DECLARE_DEVICE_TYPE(PDS030_XCEEDMC30, device_nubus_card_interface)
DECLARE_DEVICE_TYPE(PDS030_XCEEDMC30INT, device_nubus_card_interface)

#endif // MAME_BUS_NUBUS_PDS30_30HR_H
