// license:BSD-3-Clause
// copyright-holders:R. Belmont
/*********************************************************************

    Sirius JoyPort

*********************************************************************/

#ifndef MAME_BUS_A2GAMEIO_JOYPORT_H
#define MAME_BUS_A2GAMEIO_JOYPORT_H

#pragma once

#include "bus/a2gameio/gameio.h"

// device type declaration
DECLARE_DEVICE_TYPE(APPLE2_JOYPORT, device_a2gameio_interface)

#endif // MAME_BUS_A2GAMEIO_JOYPORT_H
