// license:BSD-3-Clause
// copyright-holders:Nigel Barnes
#ifndef MAME_BUS_ELECTRON_SIDEWNDR_H
#define MAME_BUS_ELECTRON_SIDEWNDR_H

#pragma once

#include "exp.h"


// device type definition
DECLARE_DEVICE_TYPE(ELECTRON_SIDEWNDR, device_electron_expansion_interface)

#endif // MAME_BUS_ELECTRON_SIDEWNDR_H
