// license:BSD-3-Clause
// copyright-holders:Vas Crabb
/***************************************************************************

 Hudson Soft HuC-3 Memory Controller

 ***************************************************************************/
#ifndef MAME_BUS_GAMEBOY_HUC3_H
#define MAME_BUS_GAMEBOY_HUC3_H

#pragma once

#include "slot.h"


DECLARE_DEVICE_TYPE(GB_ROM_HUC3, device_gb_cart_interface)

#endif // MAME_BUS_GAMEBOY_HUC3_H
